#include "include/wingetscreen/wingetscreen_plugin.h"

// This must be included before many other Windows headers.
#include <windows.h>
#include <iostream>
#include <fstream>
// For getPlatformVersion; remove unless needed for your plugin implementation.
#include <VersionHelpers.h>
#include <flutter/method_channel.h>
#include <flutter/plugin_registrar_windows.h>
#include <flutter/standard_method_codec.h>
#include <map>
#include <memory>
#include <sstream>
#include <vector>
#include <functional>
#include <mutex>
#include <condition_variable>
#include <d3d11.h>
#include <d3d11_4.h>
#include <string>
using std::string;


#include <pshpack2.h>
struct BitmapHeader
{
    unsigned short bfType;
    unsigned long bfSize;
    unsigned short bfReserved1;
    unsigned short bfReserved2;
    unsigned long bfOffBits;
    unsigned int biSize;
    int biWidth;
    int biHeight;
    unsigned short biPlanes;
    unsigned short biBitCount;
    unsigned int biCompression;
    unsigned int biSizeImage;
    int biXPixPerMeter;
    int biYPixPerMeter;
    unsigned int biClrUsed;
    unsigned int biClrImporant;
};
#include <poppack.h>
#include <cassert>


namespace {

class WingetscreenPlugin : public flutter::Plugin {
 public:
  static void RegisterWithRegistrar(flutter::PluginRegistrarWindows *registrar);

  WingetscreenPlugin();

  virtual ~WingetscreenPlugin();

 private:
  // Called when a method is called on this plugin's channel from Dart.
  void HandleMethodCall(
      const flutter::MethodCall<flutter::EncodableValue> &method_call,
      std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
};


void base64Encode(char* src, char* dtc, int len, int* dtc_len) {
    //
    //                              1         2         3         4         5         6
    //                    0123456789012345678901234567890123456789012345678901234567890123456789
    //                                    1               2               3
    //                    0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF
    const char table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

    int n;
    int mod = len % 3;

    int adj_len = len - mod;
    char* p = dtc;
    int o0, o1, o2, o3;
    for (n = 0; n < adj_len;) {
        unsigned x = ((unsigned)src[n] << 16) + ((unsigned)src[n + 1] << 8) + unsigned(src[n + 2]);
        o0 = (x >> 18) & 0x3f;
        o1 = (x >> 12) & 0x3f;
        o2 = (x >> 6) & 0x3f;
        o3 = x & 0x3f;
        *p++ = table[o0];
        *p++ = table[o1];
        *p++ = table[o2];
        *p++ = table[o3];
        n += 3;
    }
    if (mod) {
        if (mod == 1) {
            unsigned x = (unsigned)src[n] << 16;
            o0 = (x >> 18) & 0x3f;
            o1 = (x >> 12) & 0x3f;
            *p++ = table[o0];
            *p++ = table[o1];
            *p++ = '=';
            *p++ = '=';
        }
        else if (mod == 2) {
            unsigned x = ((unsigned)src[n] << 16) + ((unsigned)src[n + 1] << 8);
            o0 = (x >> 18) & 0x3f;
            o1 = (x >> 12) & 0x3f;
            o2 = (x >> 6) & 0x3f;
            *p++ = table[o0];
            *p++ = table[o1];
            *p++ = table[o2];
            *p++ = '=';
        }
    }
    *p = 0;
    *dtc_len = int(p - dtc);
}



using BitmapCallback = std::function<void(const void* data, int width, int height)>;




template<class Blt>
bool CaptureImpl(RECT rect, HWND hwnd, const BitmapCallback& callback, const Blt& blt)
{
    int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;

    BITMAPINFO info{};
    info.bmiHeader.biSize = sizeof(info.bmiHeader);
    info.bmiHeader.biWidth = width;
    info.bmiHeader.biHeight = height;
    info.bmiHeader.biPlanes = 1;
    info.bmiHeader.biBitCount = 32;
    info.bmiHeader.biCompression = BI_RGB;
    info.bmiHeader.biSizeImage = width * height * 4;

    bool ret = false;
    HDC hscreen = ::GetDC(hwnd);
    HDC hdc = ::CreateCompatibleDC(hscreen);
    void* data = nullptr;
    if (HBITMAP hbmp = ::CreateDIBSection(hdc, &info, DIB_RGB_COLORS, &data, NULL, NULL)) {
        ::SelectObject(hdc, hbmp);
        blt(hscreen, hdc);
        callback(data, width, height);
        ::DeleteObject(hbmp);
        ret = true;
    }
    ::DeleteDC(hdc);
    ::ReleaseDC(hwnd, hscreen);
    //saveBMP("DIBSection32_bmp24.bmp", (LPBYTE)data, &info);//save bmp

    //printf("%s", tmp);
    return ret;
}


int ToBitmapStep(int step)
{
    int paddings[] = { 0, 3, 2, 1 };
    return step + paddings[step % 4];
}


void SaveBitmap24(std::string path, int width, int height, const unsigned char* pixels)
{
    std::ofstream file(path, std::ios::binary);
    assert(file.is_open());
    int step = ToBitmapStep(3 * width);

    BitmapHeader header;
    header.bfType = 0x4d42; // "BM"
    header.bfSize = step * height + 54;
    header.bfReserved1 = 0;
    header.bfReserved2 = 0;
    header.bfOffBits = 54;
    header.biSize = 40;
    header.biWidth = width;
    header.biHeight = height;
    header.biPlanes = 1;
    header.biBitCount = 24;
    header.biCompression = 0;
    header.biSizeImage = 0;
    header.biXPixPerMeter = 0;
    header.biYPixPerMeter = 0;
    header.biClrUsed = 0;
    header.biClrImporant = 0;
    file.write((char*)&header, sizeof(header));

    std::vector<unsigned char> bmpPixels(step * height);

    //for (int i = 0; i < height; ++i)
    //    memcpy(bmpPixels.data() + i * step, pixels + i * 3 * width, 3 * width);
    for (int i = 0; i < height; i++) {
        for (int j = 0; j < width; j++) {
            memcpy(bmpPixels.data() + i * width * 3 + j*3,   pixels + (height  -1 - i) * width * 4 + j*4+2, 1);
            memcpy(bmpPixels.data() + i * width * 3 + j*3+1, pixels + (height - 1 - i) * width * 4 + j*4+1, 1);
            memcpy(bmpPixels.data() + i * width * 3 + j*3+2, pixels + (height - 1 - i) * width * 4 + j*4  , 1);
        }
    }
    file.write((char*)bmpPixels.data(), bmpPixels.size());
    return;
}


bool SaveAsPNG(const char* path, int w, int h, int src_stride, const void* data, bool flip_y)
{
    //return false; // for profile
    std::cout << w<<":"<<h<<"\n";
    std::vector<byte> buf(w * h * 4);
    int dst_stride = w * 4;
    auto src = (const byte*)data;
    auto dst = (byte*)buf.data();
    if (flip_y) {
        for (int i = 0; i < h; ++i) {
            auto s = src + (src_stride * (h - i - 1));
            auto d = dst + (dst_stride * i);
            for (int j = 0; j < w; ++j) {
                d[0] = s[2];
                d[1] = s[1];
                d[2] = s[0];
                d[3] = s[3];
                s += 4;
                d += 4;
            }
        }
    }
    else {
        for (int i = 0; i < h; ++i) {
            auto s = src + (src_stride * i);
            auto d = dst + (dst_stride * i);
            for (int j = 0; j < w; ++j) {
                d[0] = s[2];
                d[1] = s[1];
                d[2] = s[0];
                d[3] = s[3];
                s += 4;
                d += 4;

            }
        }
    }
    SaveBitmap24("sss.bmp", w, h, buf.data());
    //return stbi_write_png(path, w, h, 4, buf.data(), dst_stride);
    //savebmp(path, w, h, 4, buf.data(), dst_stride);
    std::cout << buf.size();
    return true;
}


bool CaptureMonitor(HMONITOR hmon, const BitmapCallback& callback)
{

    MONITORINFO info{};
    info.cbSize = sizeof(MONITORINFO);
    ::GetMonitorInfoA(hmon, &info);

    RECT rect = info.rcMonitor;
    int x = rect.left;
    int y = rect.top;
    int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;

    return CaptureImpl(rect, nullptr, callback, [&](HDC hscreen, HDC hdc) {
        ::StretchBlt(hdc, 0, 0, width, height, hscreen, x, y, width, height, SRCCOPY);
        });
}


void TestGDI()
{
    HMONITOR target = ::MonitorFromPoint({ 0, 0 }, MONITOR_DEFAULTTOPRIMARY);
    CaptureMonitor(target, [](const void* data, int w, int h) {
        //std::cout << data << "\n";
        SaveAsPNG("CaptureMonitor.png", w, h, w * 4, data, true);
        std::cout << "captured\n";
        });
    std::cout << "testgdiend\n";
}




// static
void WingetscreenPlugin::RegisterWithRegistrar(
    flutter::PluginRegistrarWindows *registrar) {
  auto channel =
      std::make_unique<flutter::MethodChannel<flutter::EncodableValue>>(
          registrar->messenger(), "wingetscreen",
          &flutter::StandardMethodCodec::GetInstance());

  auto plugin = std::make_unique<WingetscreenPlugin>();

  channel->SetMethodCallHandler(
      [plugin_pointer = plugin.get()](const auto &call, auto result) {
        plugin_pointer->HandleMethodCall(call, std::move(result));
      });

  registrar->AddPlugin(std::move(plugin));
}

WingetscreenPlugin::WingetscreenPlugin() {}

WingetscreenPlugin::~WingetscreenPlugin() {}

void WingetscreenPlugin::HandleMethodCall(//main
    const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
  //
  if (method_call.method_name().compare("getPlatformVersion") == 0) {
    std::ostringstream version_stream;
    version_stream << "Windows ";
    if (IsWindows10OrGreater()) {
      version_stream << "10+";
    } else if (IsWindows8OrGreater()) {
      version_stream << "8";
    } else if (IsWindows7OrGreater()) {
      version_stream << "7";
    }
    result->Success(flutter::EncodableValue(version_stream.str()));
  } else if (method_call.method_name().compare("getRawBMPFile") == 0){
    //result->NotImplemented();fails
    TestGDI();
    result->Success(flutter::EncodableValue("done"));
  }else{
   result->Success(flutter::EncodableValue("command not found"));
  }
}

}  // namespace

void WingetscreenPluginRegisterWithRegistrar(
    FlutterDesktopPluginRegistrarRef registrar) {
  WingetscreenPlugin::RegisterWithRegistrar(
      flutter::PluginRegistrarManager::GetInstance()
          ->GetRegistrar<flutter::PluginRegistrarWindows>(registrar));
}
